;============================================================================
;	STM_CMD.ASM
; 
;	STM Command Interpreter
;	Copyright (C) Jrgen Mller, juergen@e-basteln.de
;
;	jm  6.11.99	number input (getnum) 
;	jm  8.11.99	command line processing (getline)
;	jm 27.11.99	remove shadow variables (now local to ssi module);
;			added more motor cmmands
;	jm 12.12.99	added noise monitoring variables, removed feedback P
;	jm 5.1.2000	new variables MB, ZL, ZH (to control Z approach and feedback);
;			fixed processing of multiple assignments/line
;	jm 7.1.2000	new variables for tilt compensation and Z step test signal
;
;	To do:
;	Rounding errors in getnum;
;	extend getnum to hex input format;
;	integer output (hex/dec)?
;
;============================================================================

DIGITS		equ	6		; significant digits in frac output

;----------------------------------------------------------------------------
; Assigned variable and command names.
; Entries must be in uppercase, exactly 2 alpha characters.
; Order of entry must be repeated in declaration order of active variable set!
;----------------------------------------------------------------------------

		org	y:		; "slow" variables

varnames	; stepper motor operation
		dc	'MP'		; current motor position
		dc	'MD'		; motor destination
		dc	'MB'		; backup distance (retract from sample, incl backlash!)

		; scan generation and sampling
		dc	'XO'		; offset (image center) X
		dc	'YO'		; offset (image center) Y
		dc	'XL'		; image size X	
		dc	'YL'		; image size Y
		dc	'XS'		; sampling increment X	
		dc	'YS'		; sampling increment Y
		dc	'XI'		; scanning increment X	
		dc	'YI'		; scanning increment Y
		dc	'XR'		; rotation factor X. XR+YR=1	
		dc	'YR'		; rotation factor Y
		dc	'XM'		; max. output change X	
		dc	'YM'		; max. output change Y

		; tunneling junction and Z feedback
		dc	'UB'		; bias voltage
		dc	'IT'		; tunneling current setpoint
		dc	'FI'		; feedback I factor
		dc	'FT'		; feedback lowpass time constant
		dc	'ZL'		; low piezo Z limit (retract when below limit)
		dc	'ZH'		; high piezo Z limit (extend up to here on approach)
		dc	'ZU'		; delta_z piezo retraction upon tip crash
		dc	'ZD'		; delta_z piezo advance upon approach
		dc	'ZS'		; Z slope (average delta Z per SSI frame)
		dc	'ZT'		; Z tilt compensation automatically?

		; monitoring and test functions
		dc	'TA'		; tunnel voltage average
		dc	'TN'		; tunnel voltage noise
		dc	'TZ'		; piezo Z pos
		dc	'TS'		; test step size (piezo Z)
		dc	'TP'		; half period of test step (# SSI frames)

VARLEN		equ	*-varnames	; number of variables declared

cmdnames	; command names, must directly follow variable name table!
		dc	'SC'		; accept new variable settings and start scan
		dc	'ZA'		; start automatic approach
		dc	'ZR'		; retract tip
		dc	'CC'		; clear crash indicator
		dc	'MG'		; motor: goto destination
		dc	'ML'		; motor: goto limit switch
		dc	'MS'		; motor: stop immediately
		dc	'MU'		; motor: up one step
		dc	'MM'		; motor: down one step
		dc	'MO'		; motor: power on
		dc	'MF'		; motor: power off
CMDLEN		equ	*-cmdnames

		dc	0		; end marker

;----------------------------------------------------------------------------
;   Variable storage
;----------------------------------------------------------------------------

		org	x:		; "fast" variables

activeset				; actual variables used internally
cmd_mp		dc	0
cmd_md		dc	0
cmd_mb		dc	600

cmd_ssi_start				; variables copied to ssi set start here
cmd_xo		dc	0.0
cmd_yo		dc	0.0
cmd_xl		dc	0.1
cmd_yl		dc	0.1
cmd_xs		dc	0.01
cmd_ys		dc	0.01
cmd_xi		dc	0.01
cmd_yi		dc	0.01
cmd_xr		dc	0.707		;0.8666		; cos phi (image rotation)
cmd_yr		dc	-0.707		;0.5		; sin phi	
cmd_xm		dc	0.0001
cmd_ym		dc	0.0001
cmd_ssi_end

cmd_ub		dc	0.015		; U_bias, 0.1 div/V
cmd_it		dc	0.0125		; I_tunnel setpoint, 0.0125 div/nA
cmd_fi		dc	0.0001		; integral feedback amplitude
cmd_ft		dc	0
cmd_zl		dc	-0.9		; low piezo Z limit (retract when below limit)
cmd_zh		dc	0.1		; high piezo Z limit (extend up to here on approach)
cmd_zu		dc	-0.003		; retraction upon tip crash (0.003 is approx 10 Angstrom?)
cmd_zd		dc	0.0003		; step size for approach (0.0003 is approx 1 Angstrom?)
cmd_zs		dc	0.0		; Z slope (average delta Z per SSI frame)
cmd_zt		dc	0		; flag: automatic tilt compensation?

cmd_ta		dc	0.0		; average tunnel current
cmd_tn		dc	0.0		; mean square fluctuations of tunnel current
cmd_tz		dc	0.0		; piezo Z position
cmd_ts		dc	0.0		; test step size (piezo Z)
cmd_tp		dc	200		; half period of test step (# SSI frames)

		IF	(*-activeset)!=VARLEN
		FAIL	'jm: Number of global variables (cmd set) does not match.'
		ENDIF

		IF	(cmd_ssi_end-cmd_ssi_start)!=SSI_VARLEN
		FAIL	'jm: Number of global variables (ssi set) does not match.'
		ENDIF

;----------------------------------------------------------------------------
;   Address table for commands
;----------------------------------------------------------------------------

		org	ph:

cmdtable	jmp	>ssi_startscan
		jmp	>ssi_approach
		jmp	>ssi_retract
		jmp	>ssi_clear
		jmp	>mot_goto
		jmp	>mot_limit
		jmp	>mot_stop
		jmp	>mot_plus
		jmp	>mot_minus
		jmp	>mot_on
		jmp	>mot_off

		IF	(*-cmdtable)!=2*CMDLEN
		FAIL	'jm: Number of commands (cmdtable) does not match.'
		ENDIF

;---------------------------------------------------------------------------
;
;      DSP initialization
;
;---------------------------------------------------------------------------
	org     ph:

cmd_init
	; nothing yet
	; ### initialize variables to defaults?
	rts

; Flush input buffer until CR found.
; A1 is expected to contain first input character.
; Modifies R1,M1,X0,X1,A

cmd_flush
	move	#>CR,x1
	jmp	>_check
_loop	jsr	sci_read
_check	cmp	x1,a
	jne	_loop
	rts

; Read from input buffer until non-whitespace char is found.
; SPACE and all control chars except CR considered as white space.
; Returns character in A1.
; Modifies R1,M1,X0,A

cmd_read
	jsr	sci_read

; Skip white space (read more chars from input buffer if required).
; First character to consider is in A

skip_space
	move	#>CR,x0
	cmp	x0,a
	jeq	_done
	move	#>' ',x0
	cmp	x0,a
	jle	cmd_read
_done	rts

; Print simple "error message" (?<CR>)
; Modifies ###

cmd_err
	putchar	#'?'
	putchar	#BEL
	putchar	#CR
	rts

; Get variable name from A1 (first char) and input buffer.
; Var names consist of 2 letters; not case sensitive.
; Returns index 0..VARLEN-1 in R0; returns -1 if name not assigned.
; Returns CS when name is invalid; 
; in that case, returns offending character (masked with $5F) in A1
; 
; Modifies R0,N0,M0,R1,M1,X0,X1,A,B

cmd_getvar
	; read to alpha chars into A1, B1
	jsr	_alpha
	jcs	<_err
	move	a,b
	jsr	sci_read
	jsr	_alpha
	jcs	<_err

	; search name in table
	move	#0,r0
	move	#varnames,n0
	move	#$FFFF,m0
	rep	#8
	lsl	b1
	add	a,b
	rep	#8
	lsl	b1
_search	move	y:(r0+n0),a
	tst	a	(r0)+
	jeq	<_end
	cmp	a,b
	jne	_search

	; found it: index+1 is in r0, Carry clear from last compare
	lua	(r0)-,r0
	rts

	; name not found: set r0 = -1, Carry is clear
_end	move	m0,r0
_err	rts

	; convert alphabetic character in a1 to 0..25
	; not case sensitive;
	; return CS when non-alpha character found.

_alpha	move	#>$5F,x0		; convert to uppercase (leaves control chars ok)
	and	x0,a1
	move	#>'A',x0
	cmp	x0,a
	jcs	_error
	move	#>('Z'+1),x0
	cmp	x0,a
	bchg	#0,SR			; invert C bit
_error	rts

;--------------------------------------------------------------------
;   Get decimal number from serial input.
;   Expects complete numeral string in input buffer;
;   stops when non-numeric character found or buffer empty.
;   Returns number in B1, character causing termination in A1.
;   Returns CS for invalid input.
;   Modifies R1,R2,M1,X0,X1,Y1,A
;--------------------------------------------------------------------

cmd_getnum

	; local status bits
_bfirst	equ	0
_bminus	equ	1
_bdec	equ	2
_bdig	equ	3
_bbad	equ	4

	clr	b			; clear result accumulator
	move	#0,r2			; clear status bits
	bset	#_bbad,r2		; input not valid yet
	move	#$666666,y1		; initial multiplier for frac conversion (0.1*2^22*16)
					; (factor 16 provides 4 extra bits for rounding) 
	jsr	cmd_read		; get non-whitespace char
	jmp	<_enter

_repeat	jsr	sci_read		; get char -> a1
_enter	move	#>' ',x1
	cmp	x1,a
	jle	<_done			; quit on any control chars
	
	bset	#_bfirst,r2		; is this the first significant character?
	jcs	<_notfirst		; no, minus sign not allowed any more
	move	#>'-',x1
	cmp	x1,a
	jne	<_notfirst
	bset	#_bminus,r2		; remember the '-'
	jmp	_repeat

_notfirst
	bset	#_bdig,r2		; has any digit or decimal been processed?
	jcs	<_nodecimal		; yes, can't use decimal point any more
	move	#>'.',x1
	cmp	x1,a			; decimal point?
	jne	<_nodecimal
	bset	#_bdec,r2		; remember the '.'
	jmp	_repeat

_nodecimal
	move	#>'9',x1
	cmp	x1,a
	jgt	<_done			; invalid character
	move	#>'0',x1		; convert to decimal (0..9)
	cmp	x1,a
	jlt	<_done			; invalid character -- terminate
	sub	x1,a
	bclr	#_bbad,r2		; first "real" digit -- valid number now
	jset	#_bdec,r2,_isfrac	; fractional number

	; convert integer...
	move	b1,x1			; b1 = 10*b1 + a1
	lsl	b
	lsl	b
	add	x1,b
	lsl	b
	add	a,b
	jmp	_repeat

	; convert fraction
_isfrac	move	a1,x1
	mac	x1,y1,b   #.1,x1	; number is assembled in b0, but shifted left 4 bits!
	mpy	x1,y1,a
	move	a1,y1
	jmp	_repeat
		
_done	jclr	#_bdec,r2,_isint	; integer: no post processing

	; fraction post processing
	rep	#20			; shift result into position
	norm	r1,b
	rnd	b

_isint	jclr	#_bminus,r2,_ispos
	neg	b

_ispos	btst	#_bbad,r2		; set carry if input invalid
	rts


;--------------------------------------------------------------------
;   Write fractional number to serial output buffer
;   Expects number in B1.
;   Modifies R1,M1,X0,Y0,Y1,A,B
;--------------------------------------------------------------------

cmd_putfrac

	; handle negative numbers
	tst	b
	jpl	<_pos
	neg	b
	move	#'-',a1
	jsr	sci_write

_pos	move	#'.',a1
	jsr	sci_write
	move	#>10,y0			; multiplier
	move	#>'0',y1		; offset for ASCII conversions

	; convert digits: fixed number of significant dig's
	do	#DIGITS,_loop
	move	b1,x0
	mpy	y0,x0,a
	move	a0,b1
	add	y1,a
	jsr	sci_write
	lsr	b
_loop	rts

;--------------------------------------------------------------------
;   Process one command contained in serial input buffer.
;   Returns after execution of one command to keep down processing time
;   (multiple commands may be on one CR-terminated line)
;   Modifies R0,N0,M0,R1,M1,R2,X0,X1,Y1,A,B
;--------------------------------------------------------------------

cmd_getline

_loop	jsr	cmd_read		; get first char

_loop1	jsr	skip_space		; skip white space
	move	#>CR,x1			; next character in A
	cmp	x1,a
	jeq	<_done
	jsr	cmd_getvar		; get variable 
	jcs	<_error			; bad character in name
	btst	#15,r0
	jcs	<_error2		; negative index: name not defined
	jsr	cmd_read		; read command char, skip whitespace

	; determine command 
	move	#>'=',x1
	cmp	x1,a
	jeq	<_assign		; "=": assignment

	move	#>'?',x1
	cmp	x1,a
	jeq	<_read			; "?": read out variable

	move	#>'!',x1
	cmp	x1,a
	jne	<_error			; "!": triggers action	

	; action
	move	r0,b	
	move	#>VARLEN,x0
	sub	x0,b
	jcs	<_error			; variable name specified instead of command
	lsl	b
	move	b,r0
	move	#cmdtable,n0		; address lookup
	nop
	jsr	(r0+n0)
	jmp	_loop			; continue, get next character

	; assignment
_assign	move	r0,b	
	move	#>VARLEN,x0
	cmp	x0,b
	jge	<_error			; command name specified instead of variable
	move	r0,y:(r6)+		; save variable index
	jsr	cmd_getnum		; get numeric constant
	move	y:-(r6),r0
	move	#activeset,n0
	jcs	<_error			; invalid number in cmd_getnum?
	move	b,x:(r0+n0)
	jmp	_loop1			; continue, next character in A1

	; read variable
_read	move	r0,b	
	move	#>VARLEN,x0
	cmp	x0,b
	move	#activeset,n0		; readout refers to active variables
	jge	<_error			; command name specified instead of variable
	move	x:(r0+n0),b
	jsr	cmd_putfrac
	putchar	#CR
	jmp	_loop			; continue, get next character

_done	rts

	; error (invalid command)
_error2	jsr	sci_read
_error	jsr	cmd_flush		; invalid command: flush serial buffer
	jmp	cmd_err

