;============================================================================
;	STM_SSI.ASM
; 
;	Driver for synchronuous serial interface,
;	piezo X/Y scan control and Z approach and feedback.
;
;	Copyright (C) Jrgen Mller, juergen@e-basteln.de
;
;	Hardware configuration:
;	CS4215 codec; 2 MAX542 DACs and 1 MAX195 ADC.
;
;	jm  8.11.99	derived from sawtooth.asm, ada_8.asm;
;			adapted to 60 MHz processor clock
;	jm 14.11.99	x/y scan implemented (incl. rotation, scaling, offset)
;	jm 16.11.99	limited slope for Y return (uses ssi_xi steps)
;	jm 12.12.99	monitor for tunnel current noise
;	jm 4.1.2000	logarithm function
;	jm 5.1.2000	feedback loop and Z approach
;	jm 7.1.2000	step test signal for feedback tuning; 
;			tilt compensation and automatic adaptation
;	jm 16.1.2000	new order of time slots to reduce feedback loop delay.
;			Z-output moved from time slot 5 to 7, allows resonse in same frame.
;			Requires STM_ADDA2.GAL!
;	jm 24.4.2000	fixed ssi_slot macro (no more spurious returns)
;	jm 2.9.2000	added FEEDBACKTEST mode (simple integrative response)
;
;	To do (optional):
;	- Make frame sync routine an IRQ A handler?
;	- Make rx/tx fast interrupts? (reserve R3, R4; re-sync in frame sync handler)
;	- Add tilt compensation (better tracking by "guessing" tilt contrib. in advance)
;
;	Notes: 
;	- Overwrites PCC! Initialize SCI after SSI module!
;	- Enables interrupts! 
; 
;============================================================================

; Sign convention for piezo Z:
; Positive DA control word --> Piezo expands
; If the hardware behaves differently, set Z_INVERT flag!

Z_SEPARATE	equ	1		; 0 = Z signal is added to X/Y electrodes (flag controls approach limits)
Z_INVERT	equ	Z_SEPARATE	; inverse polarity of piezo Z when using inner electrode

; Test cases (flags control IF directives)

DELAYTEST	equ	0		; echo tunnel input as Z output directly for pass-through delay measurment
FEEDBACKTEST	equ	0		; echo integrated tunnel input as Z output for frequency response measurement

;--------------------------------------------------------------------------
;	Modified ada_init code from EVM56002 kit; 
;	some parts are Copyright (c) MOTOROLA 1995, 
;	Semiconductor Products Sector, DSP Division.
;--------------------------------------------------------------------------
;
;	port C usage:
;
;     	bit8: SSI TX (from DSP to Codec)
;	bit7: SSI RX (from Codec to DSP)
;	bit6: SSI Clock
;	bit5: SSI Frame Sync
;	bit4: codec reset (from DSP to Codec)
;	bit3:
;     	bit2: 0=control, 1=data
;
;--------------------------------------------------------------------------
;
;  Defined Constants used to construct data for the CONTROL TIME SLOTS

NO_PREAMP       equ     $100000 ;0 == enable 20 dB pre-amp 
LO_OUT_DRV      equ     $080000 ;0 == 2.8 Vp-p line (1V rms)
                                ;0 == 4.0 Vp-p headphones
                                ;1 == Line and Headphone 2.0Vp-p
HI_PASS_FILT    equ     $008000 ;0 == HPF disabled
SAMP_RATE_9     equ     $003800 ; 9.6 kHz sample rate
SAMP_RATE_48    equ     $003000 ;48   kHz sample rate
SAMP_RATE_32    equ     $001800 ;32   kHz sample rate
SAMP_RATE_27    equ     $001000 ;27.4 kHz sample rate
SAMP_RATE_16    equ     $000800 ;16   kHz sample rate
SAMP_RATE_8     equ     $000000 ; 8   kHz sample rate
STEREO          equ     $000400 ;1 == stereo, 0 == mono
DATA_8LIN       equ     $200300 ; 8-bit unsigned linear
DATA_8A         equ     $200200 ; 8-bit A-law
DATA_8U         equ     $200100 ; 8-bit u-law
DATA_16         equ     $200000 ;16-bit 2s complement linear
IMMED_3STATE    equ     $800000 ;1=SCLK & FS 3-state immediately
XTAL2_SELECT    equ     $200000 ;REQUIRED as this is the only clock
                                ;  source on the board
BITS_64         equ     $000000 ; 64 bits per frame
BITS_128        equ     $040000 ;128 bits per frame
BITS_256        equ     $080000 ;256 bits per frame
CODEC_MASTER    equ     $020000 ;1 == codec generates SCLK & FS
                                ;0 == codec receives  SCLK & FS
CODEC_TX_OFF    equ     $010000 ;0 == enable codec TX to DSP
                                ;1 == disable (Hi-Z) codec output

;--------------------------------------------------------------------------
;
;  Defined Constants used to construct data for the DATA TIME SLOTS  (5-8)

HEADPHONE_EN    equ     $800000 ;1 == headphone output enabled, 0 == muted        
LINEOUT_EN      equ     $400000 ;1 == line      output enabled, 0 == muted
LEFT_ATTN       equ     $010000 ;63 steps * 1.5 dB = -94.5 dB
SPEAKER_EN      equ     $004000 ;1 == speaker   output enabled, 0 == muted
RIGHT_ATTN      equ     $000100 ;63 steps * 1.5 dB = -94.5 dB
MIC_IN_SELECT   equ     $100000 ;1 == A/D inputs from MIC pins  NOTE: the
                                ;   DSP56002EVM uses these pins.  The line
                                ;   input pins are not used.
LEFT_GAIN       equ     $010000 ;15 steps * 1.5 dB =  22.5 dB
MONITOR_ATTN    equ     $001000 ;15 steps * 6.0 dB =  90.0 dB (mute)
RIGHT_GAIN      equ     $000100 ;15 steps * 1.5 dB =  22.5 dB

;----------------------------------------------------------------------------
;
;  Constructed constants for codec set up/initialize

CTRL_WD_12      equ     NO_PREAMP+LO_OUT_DRV+SAMP_RATE_32+STEREO+DATA_16
					; CLB=0
					; jm 21.12.98: OLB=1, NO HIGHPASS. 
CTRL_WD_34      equ     IMMED_3STATE+XTAL2_SELECT+BITS_128+CODEC_MASTER
				; jm 26.11.98: BITS_128, RATE_32
CTRL_WD_56      equ     $000000
CTRL_WD_78      equ     $000000

;----------------------------------------------------------------------------
;
;  Constructed constants for codec data mode

OUTPUT_SET      equ     HEADPHONE_EN+LINEOUT_EN+(LEFT_ATTN*4)
INPUT_SET       equ     MIC_IN_SELECT+(15*MONITOR_ATTN)+(RIGHT_ATTN*4)

;----------------------------------------------------------------------------
; The two buffers which are defined below are the source and
; destination storage for the codec Input/Output ISRs.
;
; jm 20.12.98  doubled rx and tx buffer length
;----------------------------------------------------------------------------
       
		org    x:		; fast internal RAM

RX_BUFF_BASE    dsm	8
TX_BUFF_BASE    dsm	8

RX_PTR          ds      1       	; Pointer for rx buffer
TX_PTR          ds      1       	; Pointer for tx buffer

tx_x		equ	1		; used rx/tx time slots
tx_y		equ	0
rx_it		equ	4
tx_ub		equ	4
tx_z		equ	7

;----------------------------------------------------------------------------
;	interrupt vectors
;----------------------------------------------------------------------------

	org     p:$000C
	jsr     ssi_rx_isr              ;SSI RX
	jsr     ssi_rx_isr              ;SSI RX w/Exception
	jsr     ssi_tx_isr              ;SSI TX
	jsr     ssi_tx_isr              ;SSI TX w/Exception

;----------------------------------------------------------------------------
;	global variables
;----------------------------------------------------------------------------

		org	x:		; fast variables

; internal variables, changed during scan

x_scan		dc	0		; x scan index (runs -1..+1 by ssi_xi for each line)
y_scan		dc	0		; y scan index (runs -1..+1 by ssi_yi for scan)
x_sample	dc	0		; x sampling index (runs -1..+1 by ssi_xs for each line)
y_sample	dc	0		; y sampling index (runs -1..+1 by ssi_ys for scan)

; parameters, changed by user
; must be declared in same order as corresponding cmd_... variables!

ssi_vars
ssi_xo		dc	0.0
ssi_yo		dc	0.0
ssi_xl		dc	0.0
ssi_yl		dc	0.0
ssi_xs		dc	0.01
ssi_ys		dc	0.01
ssi_xi		dc	0.01
ssi_yi		dc	0.01
ssi_xr		dc	0.99999		; cos phi (image rotation)
ssi_yr		dc	0.0		; sin phi	
ssi_xm		dc	0.00001
ssi_ym		dc	0.00001

SSI_VARLEN	equ	*-ssi_vars

;---------------------------------------------------------------------------
;      DSP/codec initialization
;---------------------------------------------------------------------------
	org     ph:

ssi_init

;   Initialize the CS4215 codec
;
;   1.  Configure the SSI port and the GPIO lines used (D/C~ & Reset)
;   2.  Select Control Mode and reset the codec (50ms @ 60MHz)
;   3.  Initialize the Transmit Data Buffer with Control Words (CLB=0)
;   4.  Enable SSI
;   5.  Wait until CLB in receive buffer = 0
;   6.  Set CLB = 1 in TX buffer
;   7.  Send 4 frames of data
;   8.  Disable SSI port
;   9.  Program SSI to receive Frame Sync and Clock
;  10.  Select Data Mode (D/C~ = 1)
;  11.  Enable SSI port

	move	#RX_BUFF_BASE,x0
	move	x0,x:RX_PTR             ; Initialize the rx pointer
	move	#TX_BUFF_BASE,x0
	move	x0,x:TX_PTR             ; Initialize the tx pointer

;   Initialize SSI
;
;   Assuming the DSP56002 is running at 60 MHz, this section selects a
;   5 MHz SSI clock to be sent from the 56002's SSI clock circuit 
;   to the codec.  (Twice the standard clock rate, since frames are 8 words
;   long to accomodate the additional AD/DA converters.)
;
;   It also sets up 128-bit frame length (8 words per frame x 16-bits per word), 
;   enables SSI TX & RX, enables TX & RX interrupts and synchronous operation 
;   with bit-length frame sync.


	movep   #$0000,x:PCC    	; turn off ssi port 
	movep   #$4702,x:CRA    	; 16bits,8 word/frame, 60/4/3=5 MHz SCLK
;	movep   #$4303,x:CRA    	; 40MHz/4/4 = 2.5MHz SCLK, WL=16 bits, 4W/F

	movep   #$FB30,x:CRB    	; RIE,TIE,RE,TE, NTWK, SYN, FSR/RSR->bit
	movep   #$14,x:PCDDR    	; setup pc2 and pc4 as outputs
	movep   #$0,x:PCD       	; D/C~ and RESET~ = 0 ==> control mode
					
	; reset delay for codec 
	do      #500,_delay_loop
	rep     #3000           	; 100 us delay @ 60 MHz
	nop
_delay_loop

	bset    #4,x:PCD        	; RESET~ = 1
	movep   #$01E8,x:PCC    	; Turn on ssi port

;
; set up the TX buffer with control mode data
;
	move    #CTRL_WD_12,x0
	move    x0,x:TX_BUFF_BASE
	move    #CTRL_WD_34,x0
	move    x0,x:TX_BUFF_BASE+1
	move    #CTRL_WD_56,x0
	move    x0,x:TX_BUFF_BASE+2
	move    #CTRL_WD_78,x0
	move    x0,x:TX_BUFF_BASE+3

	andi    #$FC,mr         	; enable interrupts!

;
; CLB == 0 in TX Buffer,  wait for CLB == 1 in RX Buffer
;
	jclr    #3,x:SSISR,*            ; wait until rx frame bit==1
	jset    #3,x:SSISR,*            ; wait until rx frame bit==0
	jclr    #3,x:SSISR,*            ; wait until rx frame bit==1
	jset    #18,x:RX_BUFF_BASE,*    ; loop until CLB set

;
; CLB == 1 in RX Buffer, send 4 frames and then disable SSI
;
	bset    #18,x:TX_BUFF_BASE      ; set CLB
	do      #4,_init_loopB          ; Delay as 4 full frames to pass
	jclr    #2,x:SSISR,*            ; wait until tx frame bit==1
	jset    #2,x:SSISR,*            ; wait until tx frame bit==0
_init_loopB
	movep   #0,x:PCC                ; reset SSI port (disable SSI...)

;
;    now CLB should be 1 -- re-program fsync and sclk direction to input
;
	movep   #$4702,x:CRA    	; 16bits,8 word/frame, 60/4/3=5 MHz
;	movep   #$4701,x:CRA    	; 16bits,8 word/frame, 40/4/2=5 MHz
	movep   #$FB00,x:CRB    	; rcv,xmt & int ena,netwk,syn,sclk==inp,msb 1st
	movep   #$14,x:PCD      	; D/C~ pin = 1  ==> data mode
	movep   #$01E8,x:PCC    	; turn on ssi port (enable SSI now...)

; high-level initializations

	jsr	<ssi_startscan		; start x/y scanning
	jmp	<ssi_startz		; keep tip retracted

;--------------------------------------------------------------------
;   SSI Receive ISR
;   R6 should point to the first free location on the stack (in Y:memory).
;   The data is placed in a 1 frame (4 word) buffer;
;   sync is verified/restored every frame.
;--------------------------------------------------------------------

	org	pl:

ssi_rx_isr
	;set_test	0,1		; set test pin
	move    r0,y:(r6)+              ; Save r0 to the stack.
	move    m0,y:(r6)+              ; Save m0 to the stack.
	move    #7,m0                   ; Modulus 8 buffer.	jm 20.12.98
	move    x:RX_PTR,r0             ; Load the pointer to the rx buffer.
	jclr    #3,x:SSISR,_next_rx	; If not frame sync, jump to receive data.
	move    #RX_BUFF_BASE,r0        ; If frame sync, reset base pointer.
	nop
_next_rx
	movep   x:SSIDR,x:(r0)+         ; Read out received data to buffer.
	move    r0,x:RX_PTR             ; Update rx buffer pointer.
	move    y:-(r6),m0              ; Restore m0.
	move    y:-(r6),r0              ; Restore r0.
	;set_test	0,0		; set clear pin
	rti

;--------------------------------------------------------------------
;   SSI Transmit ISR
;   R6 should point to the first free location on the stack (in Y:memory).
;   The data is taken from a 1 frame (4 word) buffer;
;   sync is verified/restored every frame.
;--------------------------------------------------------------------

ssi_tx_isr
	;set_test	0,1		; set test pin
	move    r0,y:(r6)+              ; Save r0 to the stack.
	move    m0,y:(r6)+              ; Save m0 to the stack.
	move    #7,m0                   ; Modulus 8 buffer.	jm 20.12.98
	move    x:TX_PTR,r0             ; Load the pointer to the tx buffer.
	jclr    #2,x:SSISR,_next_tx	; If not frame sync, jump to transmit data.
	move    #TX_BUFF_BASE+1,r0      ; If frame sync, reset pointer.
	nop
_next_tx
	movep   x:(r0)+,x:SSIDR         ; SSI transfer data register.
	move    r0,x:TX_PTR             ; Update tx buffer pointer.
	move    y:-(r6),m0              ; Restore m0.
	move    y:-(r6),r0              ; Restore r0.
	;set_test	0,0		; set test pin
	rti


;--------------------------------------------------------------------
;   Macro definitions
;--------------------------------------------------------------------

ssi_fsync	macro
	jset    #2,x:SSISR,*            ; Wait for frame sync to pass.
	jclr    #2,x:SSISR,*            ; Wait for frame sync.
	endm

ssi_slot	macro	n		; Wait until slot n received
	move	#>RX_BUFF_BASE+n,x0
_wait	move	x:RX_PTR,a	
	cmp	x0,a
	jne	_wait
	endm

;--------------------------------------------------------------------
;   Support functions
;--------------------------------------------------------------------

;--------------------------------------------------------------------
; Logarithm approximation, A = log (abs (A))
;
; A contains argument, and result upon return
; Modifies R0.
;
; Returns 4-bit log of most significant "1"-bit in a22..a19;
; linear interpolation in a18...
; log (0.5)  = 0 1111 000...
; log (0.25) = 0 1110 000...
; log (0.75) = 0 1111 100...

log
	abs	a			; use absolute value
	move	#15,r0			; count down power of two during normalization
	rep	#15
	norm	r0,a
	lsl	a1			; discard a23 (empty sign bit)
	lsl	a1			; discard a22 (described by power of two)
	move	r0,a2
	rep	#5			; shift power of two into bits a22..a19;
	asr	a			; leave linear approximation in bits a18..
	rts
	
;--------------------------------------------------------------------
;   Scan initialization
;--------------------------------------------------------------------

; CS4215 operation mode - transmitted via TX_BUFF

TONE_OUTPUT     EQU     LINEOUT_EN 	; +HEADPHONE_EN+(4*LEFT_ATTN)+(4*RIGHT_ATTN)   
TONE_INPUT      EQU     0		; MIC_IN_SELECT+(15*MONITOR_ATTN)
	org	ph:

;--------------------------------------------------------------------
;   Copy cmd set to active variables

ssi_copyvars	
	move	#cmd_ssi_start,r0
	move	#ssi_vars,r1
	move	#-1,m0
	move	m0,m1
	do	#SSI_VARLEN,_loop
	move	x:(r0)+,x0
	move	x0,x:(r1)+
_loop	rts

;--------------------------------------------------------------------
; scan re-initialization,
; called from cmd interpreter after parameters have changed

ssi_startscan
	jsr	<ssi_copyvars		; activate new parameters
	jsr	<ssi_newscan		; re-initialize scan

	move	x:cmd_it,a		; calculate log of I_tunnel setpoint
	jsr	log
	move	a,x:logit

	btst	#0,x:cmd_zt		; automatic tilt adaptation?
	jcc	<_return
	move	#0,x0
	move	x0,x:cmd_zs
	move	x0,x:zs0

_return	rts

;--------------------------------------------------------------------
; internal scan re-init,
; called after Y scan complete

ssi_newscan
	move	#-1.0,x0
	move	x0,x:x_scan
	move	x0,x:y_scan
	move	x0,x:x_sample
	move	x0,x:y_sample

	move	#0,x0
	bset	#restart,x0		; scan status: restart with new parameters
	move	x0,x:scanstatus

	move    #>TONE_OUTPUT,x0	; headphones, line out, mute spkr, no attn.
	move    x0,x:TX_BUFF_BASE+2
	move    #>TONE_INPUT,x0		; no input gain, monitor mute
	move    x0,x:TX_BUFF_BASE+3

	rts

;--------------------------------------------------------------------
;   Z control (combined motor and piezo approach/retract)
;--------------------------------------------------------------------

;--------------------------------------------------------------------
; Inititalize to passive mode
; (tip fully retracted)

ssi_startz
	move	#0,x0
	move	x0,x:zstatus
	rts

;--------------------------------------------------------------------
; Automated Z approach, using motor and piezo.
;
; Sets appropriate flags only; 
; approach is done in timer interrupt and SSI frame sync handlers.

ssi_approach
	set_led	led_motor,red
	set_led	led_tunnel,green

	bclr	#zactive,x:zstatus
	bset	#zsearch,x:zstatus
	rts

;--------------------------------------------------------------------
; Retract tip (piezo and motor):
; fully retract piezo, move motor up cmd_mb steps.
;
; Sets appropriate flags only; 
; motor and piezo control is done in timer interrupt and SSI frame sync handlers.

ssi_retract
	set_led	led_motor,yellow	; motor moving
	set_led	led_tunnel,green	; no tunnel current

	bclr	#zactive,x:zstatus	; Z loop passive
	bclr	#zsearch,x:zstatus

	move	x:cmd_mp,x0		; current motor pos
	move	x:cmd_mb,a		;   + backup distance
	abs	a
	add	x0,a
	move	a,x:cmd_md		;   = new motor destination
	motor_up			; go there
	rts

;--------------------------------------------------------------------
; Clear crash indicator LED
; Set LED to green or yellow, depending on state

ssi_clear
	btst	#zactive,x:zstatus	; Z loop active?
	jcs	<_active
	set_led	led_tunnel,green	; no tunnel current
	jmp	<_ret
_active set_led	led_tunnel,yellow	; tunnel current
_ret	rts

;--------------------------------------------------------------------
;   Scan generation and feedback loop.
;   Called after every frame sync.
;--------------------------------------------------------------------

	org	x:

; X/Y scan mode control
scanstatus	dc	0		; bitset for X/Y state machine control
x_dir		equ	0		; bit0: x scan direction is backward
restart		equ	1		; bit1: new parameters - go back to origin
xdone		equ	2		; bit2: x retrace to origin complete
ydone		equ	3		; bit3: y retrace to origin complete

; Z mode control
zstatus		dc	0		; bitset for Z state machine control
zactive		equ	0		; bit0: z feedback loop active
zsearch		equ	1		; bit1: z search mode (looking for current)
zstepsign	equ	2		; bit2: direction of next test step

; tunneling current statistics and z slope adaptation
tn0		dc	0		; storage for mean square noise estimate
tn1		dc	0
ta0		dc	0		; storage for average (low word)
zs0		dc	0		; storage for slope estimate (low word)
WEIGHT		equ	14		; 2^weight is weighting factor for sliding average

; Z feedback loop
logit		dc	0		; logarithm of I_tunnel (setpoint)
stepcount	dc	0		; count down step test period

		IF	Z_SEPARATE
ZREST		equ	-1.0		; piezo Z position at rest
		ELSE
ZREST		equ	-0.5		; piezo Z position at rest
		ENDIF

ICRASH		equ	0.94		; current that indicates tip crash
					; ### MAX195 does not read full range in current design!

	org	ph:

;--------------------------------------------------------------------
; Feedback loop

ssi_feedback

;	set_test	1,1		; set test pin


	; test for pass-through delay: echo tunneling current on Z output

	IF	DELAYTEST
	move	x:RX_BUFF_BASE+rx_it,x0	; read MAX195 (tunnel current)
	bchg	#23,x0
	move	x0,a
	asr	a
	asr	a
	bchg	#23,a1
	move	a1,x:TX_BUFF_BASE+tx_z	; output to MAX542 (z)
	jmp	<_zmon
	ENDIF

	; test for pass-through delay: echo integrated tunneling current on Z output

	IF	FEEDBACKTEST
	move	x:RX_BUFF_BASE+rx_it,x0	; read MAX195 (tunnel current)
	bchg	#23,x0
	move	x0,a
	asr	a
	asr	a
	asr	a
	move	x:TX_BUFF_BASE+tx_z,x0	; read current Z value
	bchg	#23,x0
	add	x0,a
	move	a,x0
	neg	a			; "leakage" to prevent output from drifting away
	asr	a
	asr	a
	asr	a
	asr	a
	asr	a
	asr	a
	asr	a
	asr	a
	add	x0,a
	bchg	#23,a1
	move	a1,x:TX_BUFF_BASE+tx_z	; output to MAX542 (z)
	jmp	<_zmon
	ENDIF

	; -------------------------
	; Z control
	; -------------------------

	bchg	#23,x:RX_BUFF_BASE+rx_it	; adapt from MAX195 code format
	move	x:RX_BUFF_BASE+rx_it,a	; read MAX195 (tunnel current)
	abs	a			; consider absolute value

	btst	#zactive,x:zstatus	; feedback loop active?
	jcs	<_feedback		; yes

	motor_status			; motor ready?
	jcc	<_rest			; no, wait

	move	#ICRASH,x0		; *** check for tip crash (search & idle mode)
	cmp	x0,a			; abs (I_tunnel) in A
;	jlt	<_nocrash1		; less than crash limit
	jcs	<_nocrash1		; less than crash limit
	set_led	led_tunnel,red		; indicate crash -- 
					; outside feedback mode, this is all we do
_nocrash1
	btst	#zsearch,x:zstatus	; search mode active?
	jcc	<_rest			; no, idle: output min. Z value

	; Z approach control.
	; Monitor the approach, control transition to 
	; feedback mode or next motor step.

					; *** check motor status
	set_led	led_motor,red		; indicate: approach continues
	motor_limit			; limit switch reached?
	jcc 	<_cancel		; yes, approach failed

					; *** tunneling current established?
	move	x:cmd_it,b		; get setpoint
	cmp	a,b			; setpoint reached?
	jle	<_current		; yes, switch to feedback mode

					; *** piezo fully extended?
	move	x:cmd_tz,b		; check current Z position
	move	x:cmd_zh,y0
	cmp	y0,b			; approach limit reached?  move current pos to B
	jgt	<_step			; yes, do next motor step

					; *** extend piezo further
	move	x:cmd_zd,a		; approach step size
	add	b,a			; add to current Z position
	jmp	<_zout			; new Z position in A

_step	motor_down1			; ask for next motor step
	jmp	<_rest			; goto rest position

_current
	bset	#zactive,x:zstatus	; tunneling current established:
	bclr	#zsearch,x:zstatus	; switch to feedback mode
	set_led	led_tunnel,yellow
	set_led	led_motor,green

	; Z feedback loop

_feedback				; *** check for tip crash (in feedback mode)
	move	#ICRASH,x0		; crash current
	cmp	x0,a			; abs (I_tunnel) in A
	jcs	<_nocrash2		; less than crash limit
;	jlt	<_nocrash2		; less than crash limit
	set_led	led_tunnel,red		; indicate crash

	move	x:cmd_zu,a		; delta_z correction upon crash
	abs	a
	neg	a			; make sure this is negative (retract tip)!
	jmp	<_update

_nocrash2
	jsr	log			; tunnel current readout is in A
	move	x:logit,b		; get log(setpoint)
	sub	a,b

	; Feedback sign and amplitude consideration:
	;
	; Current too large --> Z output value must be decreased.
	; Difference of logs negative --> factor cmd_fi must be positive.
	;
	; Factor 2 change in current --> difference 0.0625 in logs.
	; Required correction typ. < 1 Angstrom = 10 mV = 0.0003 div
	; (assuming 90 Angstrom/V piezo sensitivity, 40V/div D/A scale).
	; Typical cmd_fi factors <= 0.005 extected (definitely cmd_fi<1).

	move	x:cmd_fi,x0	b,y0	; get feedback amplification
	mpy	y0,x0,a
	
	; add step test signal (help for manual feedback tuning)

_update	
	move	x:stepcount,b
	dec	b			; count down step period timer
	jpl	<_nostep		; nothing due yet
	move	x:cmd_tp,b		; step: re-load timer
	move	x:cmd_ts,x0		; get step size
	bchg	#zstepsign,x:zstatus	; stepping up or down?
	jcs	<_stepup
	sub	x0,a			; step down
	jmp	<_nostep
_stepup	add	x0,a			; step up
_nostep	move	b,x:stepcount

	; add slope estimate

	move	x:cmd_zs,b		; estimated slope contribution
	btst	#x_dir,x:scanstatus
	jcc	<_forw
	neg	b			; invert if scanning backwards
_forw	add	b,a			; add to new Z position

	; update estimate (automatic slope adaptation)

	move	a,x0			; save correction term
	btst	#0,x:cmd_zt		; automatic adaptation desired?
	jcc	<_noadapt

	btst	#x_dir,x:scanstatus	; scanning backwards?
	jcc	_forw2
	neg	a			; yes, invert correction
_forw2	move	x:cmd_zs,b		; get current sum
	move	x:zs0,b0
	sub	b,a			; new - average

	; update average
	rep	#WEIGHT			; * 2^(-weight)
	asr	a
	add	a,b	
	move	b1,x:cmd_zs
	move	b0,x:zs0
_noadapt

	; retract tip if we are getting too close

	move	x:cmd_tz,a		; current Z pos. Correction is in X0
	add	x0,a
	move	x:cmd_zl,x0		; minimum permissible piezo Z pos
	cmp	x0,a
	jgt	<_zout			; still above the limit, output A

_cancel					; getting too close to surface, 
					; or unsuccessful approach (no current established)
	jsr	ssi_retract		; initiate retract sequence
_rest	move	#ZREST,a		; retract tip

_zout	move	a,x:cmd_tz		; *** common exit point with Z output
	IF Z_INVERT			; invert DA value if required by hardware
	neg	a
	ENDIF
	move	a,x0			; limit to $8000 .. $7FFF
	bchg	#23,x0			; adapt to MAX542 format
	move	x0,x:TX_BUFF_BASE+tx_z	; output to MAX542 (z)

;	set_test	1,0		; clear test pin

	; -------------------------
	; monitor tunneling current
	; -------------------------
	
_zmon	move	x:RX_BUFF_BASE+rx_it,a	; MAX195 (tunnel current); bit23 already inverted

	move	x:cmd_ta,b		; get current sum
	move	x:ta0,b0
	sub	b,a			; new - average
	move	a,x0			; remember for rms calculation

	; update average
	rep	#WEIGHT			; * 2^(-weight)
	asr	a
	add	a,b	
	move	b1,x:cmd_ta
	move	b0,x:ta0

	; update mean square
	mpy	x0,x0,a
	move	x:tn1,b			; get current sum
	move	x:tn0,b0
	sub	b,a			; new - average
	rep	#WEIGHT			; * 2^(-weight)
	asr	a
	add	a,b	
	move	b1,x:tn1
	move	b0,x:tn0
	rep	#20
	asl	b
	move	b,x:cmd_tn

	rts


;--------------------------------------------------------------------
; Scan generation

ssi_scan

	; -------------------------
	; X/Y scan
	; -------------------------

	; prepare scanning control
	move	x:x_scan,a
	move	x:y_scan,b

	; restart (move to origin) in progress?
	btst	#restart,x:scanstatus
	jcs	<_no_limit		; yes, use current x_scan and y_scan

	; increment X position
_xscan	move	x:ssi_xi,x0
	andi	#$bf,ccr		; clear limit bit
	btst	#x_dir,x:scanstatus
	jcs	<_back
	add	x0,a
	jmp	<_contx
_back	sub	x0,a
_contx	move	a,x:x_scan
	jlc	<_no_limit		; limit clear: not end of scan line

	; reverse X scan direction
	bchg	#x_dir,x:scanstatus
	jcc	<_no_limit		; going back now: same Y position
	
	; increment Y position
_conty	move	x:ssi_yi,x0
	andi	#$bf,ccr		; clear limit bit
	add	x0,b
	move	b,x:y_scan
	jlc	<_no_limit		; not end of scan

	; end of image reached
	jsr	ssi_newscan		; re-init, begin return to origin
	jmp	<_cont

_no_limit	
	; scaling and offset 
	move	a,x1			; x_scan (-1..+1)
	move	b,y1			; y_scan (-1..+1)

	move	x:ssi_xl,x0		; scaling factor X
	mpy	x0,x1,a
	move	x:ssi_xo,x0		; offset X
	add	x0,a
	move	a,x1			; rescaled X pos
	
	move	x:ssi_yl,y0		; scaling factor Y
	mpy	y0,y1,a
	move	x:ssi_yo,y0		; offset Y
	add	y0,a
	move	a,y1			; rescaled Y pos

	; rotation 
	move	x:ssi_xr,x0		; cos (phi)
	move	x:ssi_yr,y0		; sin (phi)

	mpy	x0,x1,a
	mpy	y0,y1,b
	add	b,a			; x' = x cos(phi) + y sin(phi)
	move	a,y:(r6)+		; save intended X output

	mpy	x0,y1,a
	mpy	y0,x1,b
	sub	b,a			; y' = y cos(phi) - x sin(phi)

	; check for limited slew rate mode

	btst	#restart,x:scanstatus	; restart approach?
	jcc	<_xyout			; no, regular scan
	
	btst	#ydone,x:scanstatus	; check whether approach is complete
	jcc	<_limit
	btst	#xdone,x:scanstatus
	jcc	<_limit

	; approach completed
	bclr	#restart,x:scanstatus	; done
	jmp	<_xyout			; anything else? (activate data output?)

	; limit slew rate for restart positioning
	; Y channel
	; intended Y output is in A

_limit	tfr	a,b	x:ssi_ym,x1	; x1: maximum increment allowed
	move	x:TX_BUFF_BASE+tx_y,x0	; x0: current Y position
	sub	x0,b
	jlt	<_yneg			; intended position smaller than current
	cmp	x1,b			; increment larger than slew rate limit?
	jle	<_ydone			; no, go there directly
	tfr	x0,a			; use previous pos + increment, discard stack value
	add	x1,a
	jmp	<_yout

_yneg	abs	b			; desired increment is negative
	cmp	x1,b			; abs(increment) larger than slew rate limit?
	jle	<_ydone			; no, go there directly
	tfr	x0,a			; use previous pos + increment, discard stack value
	sub	x1,a
	jmp	<_yout
	
_ydone	bset	#ydone,x:scanstatus	; y approach done
_yout	move	a,x:TX_BUFF_BASE+tx_y

	; limit slew rate for restart positioning
	; X channel
	; intended X output is on stack

	move	y:-(r6),a
	tfr	a,b	x:ssi_xm,x1	; maximum increment allowed
	move	x:TX_BUFF_BASE+tx_x,x0	; current X position
	sub	x0,b
	jlt	<_xneg			; new position smaller than previous
	cmp	x1,b			; increment larger than slew rate limit?
	jle	<_xdone			; no, go there directly
	tfr	x0,a			; use previous pos + increment, discard stack value
	add	x1,a
	jmp	<_xout

_xneg	abs	b			; desired increment is negative
	cmp	x1,b			; abs(increment) larger than slew rate limit?
	jle	<_xdone			; no, go there directly
	tfr	x0,a			; use previous pos + increment, discard stack value
	sub	x1,a
	jmp	<_xout
	
_xdone	bset	#xdone,x:scanstatus	; x approach done
_xout	move	a,x:TX_BUFF_BASE+tx_x
	jmp	<_cont


	; regular output without limiting
_xyout	move	a,x:TX_BUFF_BASE+tx_y  	; Y output
	move	y:-(r6),a
	move	a,x:TX_BUFF_BASE+tx_x	; X output

	; bias voltage output

_cont	move	x:cmd_ub,a		; bias voltage
	bchg	#23,a1			; adapt to Max542 code format
	move	a1,x:TX_BUFF_BASE+tx_ub	; MAX542 (Bias)

	rts

